unit OpenWirePinEditors;

interface

{$IFDEF VER140}
{$DEFINE D6}
{$ENDIF}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
{$IFDEF D6}
  DesignEditors,
  DesignIntf,
  TypInfo,
{$ELSE}
  dsgnintf,
{$ENDIF}
  ImgList, ComCtrls, StdCtrls, Buttons, Contnrs, OpenWirePins, ExtCtrls;

  const OWM_UPDATE              = CM_BASE + 500;
  const IIMSG_UPDATE_INSPECTOR  = CM_BASE + 501;

{$IFDEF D6}
type IADesigner = IDesigner;
type TAGetPropProc = TGetPropProc;
{$ELSE}
type IADesigner = IFormDesigner;
type TAGetPropProc = TGetPropEditProc;
{$ENDIF}

type TOWEPinType = ( ptSource, ptSink );

type
  TOWPinEntry = class
    public
      PinName           : String;
      OwnerName         : String;
      SavedChecked      : Boolean;
      Checked           : Boolean;
      CanConnect        : Boolean;
      LinkedTo          : String;
      LinkType          : String;
      Pin               : TOWPin;

  end;
  
{
  TOWESinkPinEntry = class(TOWEPinEntry)
    public
      SinkPin : TOWSinkPin;

  end;

  TOWESourcePinEntry = class(TOWEPinEntry)
    public
      SourcePin : TOWSourcePin;

  end;
}

type
  TSourcePinForm = class(TForm)
    LinksCountLabel: TLabel;
    StaticLabel: TLabel;
    OkButton: TBitBtn;
    CancelButton: TBitBtn;
    LinkAllButton: TBitBtn;
    UnlinkAllButton: TBitBtn;
    ListView: TListView;
    RestoreButton: TBitBtn;
    ImageList1: TImageList;
    ImageList2: TImageList;
    FormsComboBox: TComboBox;
    Label1: TLabel;
    Button1: TButton;
    procedure LinkAllButtonClick(Sender: TObject);
    procedure UnlinkAllButtonClick(Sender: TObject);
    procedure RestoreButtonClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure ListViewChange(Sender: TObject; Item: TListItem;
      Change: TItemChange);
    procedure ListViewColumnClick(Sender: TObject; Column: TListColumn);
    procedure FormCreate(Sender: TObject);
    procedure ListViewCompare(Sender: TObject; Item1, Item2: TListItem;
      Data: Integer; var Compare: Integer);
    procedure FormsComboBoxChange(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    { Private declarations }
    procedure UpdateLinksCount();
    procedure PopulateAll();
    procedure PopulateAllEntries();
    procedure PopulateForm( ARootComponent : TComponent );

    procedure PopulateSingleForm( CurrentRoot : TComponent; ARootComponent : TComponent; OnlyConnected : Boolean; FullPath : Boolean );
    procedure PopulateSingleFormEntries( CurrentRoot : TComponent; ARootComponent : TComponent; FullPath : Boolean );

    procedure PopulateSingleSinkForm( CurrentRoot : TComponent; ARootComponent : TComponent; OnlyConnected : Boolean; FullPath : Boolean );
    procedure PopulateSingleSourceForm( CurrentRoot : TComponent; ARootComponent : TComponent; OnlyConnected : Boolean; FullPath : Boolean );

    procedure PopulateSingleSinkFormEntries( CurrentRoot : TComponent; ARootComponent : TComponent; FullPath : Boolean );
    procedure PopulateSingleSourceFormEntries( CurrentRoot : TComponent; ARootComponent : TComponent; FullPath : Boolean );

    function  LinkToDisplayString( ARootComponent : TComponent; LinkStr : String; FullPath : Boolean; OwnRoot : Boolean; var PinStr : String ) : String;
    procedure FillFormsInfo();
    function  RootFromName( RootName : String ) : TComponent;
    
    function EntryFromPin ( Pin : TOWPin ) : TOWPinEntry;

//    function  CanAccessRootFromName( RootName : String ) : Boolean;

  private
    Designer      : IADesigner;
    SourcePin     : TOWSourcePin;
    SinkPin       : TOWSinkPin;
    Root          : TComponent;
    PinsList      : TObjectList;
    ListUpdating  : Boolean;
    
  private
    procedure OWMUpdate(var Message: TMessage); message OWM_UPDATE;
    procedure OWMUpdateInspector(var Message: TMessage); message IIMSG_UPDATE_INSPECTOR;
    
  public
    function ExecuteForSource( ADesigner : IADesigner; ASourcePin : TOWSourcePin ): Integer; virtual;
    function ExecuteForSink( ADesigner : IADesigner; ASinkPin : TOWSinkPin ): Integer; virtual;

  private
    { Private declarations }
    ColumnToSort : Integer;
    Direction    : Boolean;
    AllSelected  : Boolean;

  public
    function  GetListItems () : TListItems;
    procedure ClearData();

  public
  property Items : TListItems read GetListItems;
    { Public declarations }
  end;

procedure ResetObjectInspector( Designer : IADesigner );

procedure Register;

implementation

{$IFDEF D6}
  uses OWDesignSelectionsList, ToolsAPI, ToolIntf; //, ExptIntf;
  type TADesignerSelectionList= TOWDesignerSelectionList;
{$ELSE}
  uses ToolsAPI, ToolIntf, ExptIntf;
  type TADesignerSelectionList= TDesignerSelectionList;
{$ENDIF}

{$IFDEF BCB}
const
  SEPARATOR = '->';
{$ELSE}
const
  SEPARATOR = '.';
{$ENDIF}

const
  DISCONNECTED = '(Disconnected)';

var Form        : TSourcePinForm;
var EmptyList   : TADesignerSelectionList;
var InRefresh   : Boolean;
var InOppening  : Boolean;

type TOWModulesColection = class(TStringList)
public
  procedure GetModules( const FileName, UnitName, FormName, DesignClass: string; CoClasses: TStrings );

end;

procedure TOWModulesColection.GetModules( const FileName, UnitName, FormName, DesignClass: string; CoClasses: TStrings );
begin
  if( FormName <> '' ) then
    Add( FormName );

end;

procedure RequestRefresh();
begin
  if( InOppening ) then
    Exit;

  if( OWGetAllLinked() ) then
    Exit;

  if( PinsNeedRefresh ) then
    Exit;

  PinsNeedRefresh := True;
//  if( not PinsNeedRefresh ) then
//    Exit;

  PostMessage( Form.Handle,OWM_UPDATE, 0, 0 );
end;

function CanAccessRootFromName( Designer : IADesigner; RootName : String ) : Boolean;
var
  Component : TComponent;

begin
  Component := Designer.GetComponent( RootName + SEPARATOR + RootName );

  Result := False;
  if( Component = NIL ) then
    Exit;

  Result := Designer.IsComponentLinkable( Component );
end;
//---------------------------------------------------------------------------

procedure RequestRefreshEx( Designer : IADesigner );
var
  FormNames : TOWModulesColection;
  I         : Integer;
  
begin
  FormNames := TOWModulesColection.Create;

  Designer.GetProjectModules( FormNames.GetModules );
  for I := 0 to FormNames.Count - 1 do
    CanAccessRootFromName( Designer, FormNames.Strings[ I ] );
    
  FormNames.Free;
  
  RequestRefresh();
end;

procedure LinkAwaitsLinkingAllForms();
var
  I                     : Integer;
  ModuleIndex           : Integer;
  Module                : IOTAModule;

  ModuleServices        : IOTAModuleServices;
  FormEditor            : INTAFormEditor;

  Project               : IOTAProject;
  ModuleFileExt         : String;

begin
  if( not PinsNeedRefresh ) then
    Exit;

  if( OWGetAllLinked() ) then
    Exit;

  InOppening := True;
  PinsNeedRefresh := False;

  BorlandIDEServices.QueryInterface( IOTAModuleServices, ModuleServices );

  if( Assigned( ModuleServices )) then
    begin
    if( ModuleServices.CurrentModule.OwnerCount > 0 ) then
      begin

      Project := ModuleServices.CurrentModule.Owners[ 0 ];
//      Project._AddRef;
      for ModuleIndex := 0 to Project.GetModuleCount - 1 do
        begin
        if( Project.GetModule( ModuleIndex ).GetFileName <> '' ) then
          begin
          try
            ModuleFileExt := UpperCase( ExtractFileExt( Project.GetModule( ModuleIndex ).GetFileName )); 
            if( ( ModuleFileExt <> '.CPP' ) and ( ModuleFileExt <> '.PAS' ) and ( ModuleFileExt <> '.DFM' )) then
               Continue;
{
            if( UpperCase( ExtractFileExt( Project.GetModule( ModuleIndex ).GetFileName )) = '.RES' ) then
               Continue;

            if( UpperCase( ExtractFileExt( Project.GetModule( ModuleIndex ).GetFileName )) = '.DCP' ) then
               Continue;
}

            Project.GetModule( ModuleIndex ).OpenModule.QueryInterface( IOTAModule, Module );
            if( Project.GetModule( ModuleIndex ).FormName = '' ) then
               Continue;
               
            for I := 0 to Module.GetModuleFileCount - 1 do
              begin
                begin
//                Exit;
                Module.GetModuleFileEditor( I ).QueryInterface( INTAFormEditor, FormEditor );
                if( Assigned( FormEditor )) then
                  begin
                  OWLinkAwaitsLinking( FormEditor.GetFormDesigner.GetRoot );
                  FormEditor := NIL;
                  end;
                end;
              end;
          finally
          end
          end
        end;
      end;
    end;

  InOppening := False;
end;

procedure ResetObjectInspector( Designer : IADesigner );
begin
  if( InRefresh ) then
    Exit;
    
  if( Assigned( Designer)) then
    begin
    InRefresh := True;
    PostMessage( Form.Handle, IIMSG_UPDATE_INSPECTOR,
                Integer( Designer ), 0 );

    end;
    
end;

procedure GetSinkPinValueList( SinkPin : TOWSinkPin; List : TStrings );
begin
//  SinkPin := GetPin();
  if( SinkPin = NIL ) then
    Exit;

  OWGetPinsValueList( List, SinkPin, '.', False, False );
end;

procedure GetSinkPinValues( SinkPin : TOWSinkPin; Proc: TGetStrProc );
var
  Values : TStringList;
  i      : Integer;

begin
  Values := TStringList.Create;
  try
    GetSinkPinValueList( SinkPin, Values);
    Proc(DISCONNECTED);
    for i := 0 to Values.Count - 1 do
      Proc(Values.Strings [i]);

  except
  end;
  Values.Free;

end;

function  GetSinkPinValue( SinkPin : TOWSinkPin; Designer : IADesigner ) : string;
begin
  if( SinkPin = NIL ) then
    begin
    Result := 'Refreshing ...';
    ResetObjectInspector( Designer );
    end

  else
    begin
    Result := SinkPin.GetLinkStr( False );
    end;

end;
   
function SetSinkPinValue( Root : TComponent; SinkPin : TOWSinkPin; const Value: string ) : Boolean;
begin
  Result := False;
  if( SinkPin = NIL ) then
    Exit;

  if( SinkPin.TryLinkTo( Root, Value, False, False )) then
    begin
    Result := True;
    Exit;
    end

end;

function SetSourcePinValue( Root : TComponent; SourcePin : TOWSourcePin; const Value: string) : Boolean;
begin
  Result := False;
  if( SourcePin = NIL ) then
    Exit;

  if( SourcePin.TryLinkTo( Root, Value, False, False )) then 
    begin
    Result := True;
    Exit;
    end
    
end;

function GetSourcePinValue( SourcePin : TOWSourcePin; Designer : IADesigner ) : string;
var
  LinksCount : Integer;
  
begin
  try
    if( SourcePin = NIL ) then
      begin
      Result := 'Refreshing ...';
      ResetObjectInspector( Designer );
      Exit;
      end;

    LinksCount := SourcePin.GetSinkCountAll(); 
    if ( LinksCount = 0 ) then
      begin
      Result := '(Unlinked)';
      Exit;
      end;

    if ( LinksCount = 1 ) then
      begin
      if( SourcePin.SinkCount = 1 ) then 
        Result := SourcePin.GetLinkStr( 0, False ); //ValueToString( SourcePin.Sinks [ 0 ], '.' ); //SourcePin.Sinks [ 0 ].;
        
      Exit;
      end;

    Result := '( ' + IntToStr( LinksCount ) + ' Links )';

  except
    Result := '(Error)';

  end;

end;

procedure GetPinValueList( OwnerComponent : TComponent; Pin : TOWPin; List : TStrings );
begin
  if( Pin <> NIL ) then
    begin
    if( OWGetMainOwnerComponent( Pin.Owner ) <> OwnerComponent ) then
      OWGetPinsValueListSingleRoot( List, OwnerComponent, Pin, '.', True, OwnerComponent.Name, False )

    else
      OWGetPinsValueListSingleRoot( List, OwnerComponent, Pin, '.', True, '', False );
//    GetPinsValueList( List, SourcePin, '.', true );
    end;
end;

function GetMainDesignOwner ( Component : TComponent ) : TComponent;
begin
  if ( Component.Owner = NIL ) then
    Result := Component

  else
    Result := OWGetMainOwnerComponent ( Component.Owner );
    
end;

function SinkPinEdit( Designer : IADesigner; SinkPin : TOWSinkPin ) : Boolean;
var
  I             : Integer;
  SourcePin     : TOWSourcePin;
  SourceRoot    : TComponent;
  SourceChanged : Boolean;

begin
  Result := False;
  if( SinkPin = NIL ) then
    Exit;

  try
    if ( Form.ExecuteForSink( Designer, SinkPin ) = mrOk ) then
      begin
      for I := 0 to Form.Items.Count - 1 do
        begin
        SourceChanged := False;
        SourcePin := TOWSourcePin( Form.Items.Item [ I ].SubItems.Objects[ 0 ]);
        if ( Form.Items.Item [ I ].StateIndex = 1 ) then
          begin
          if( SinkPin.SourcePin <> SourcePin ) then
            begin
            SourceChanged := True;
            end;
          end

        else
          begin
          if( SinkPin.SourcePin = SourcePin ) then
            begin
            SourceChanged := True;
            end;
          end;

        if( SourceChanged ) then
          begin
          SourceRoot := GetMainDesignOwner( SourcePin.Owner );
          if( SourceRoot is TCustomForm ) then
            if( Assigned( TCustomForm( SourceRoot ).Designer )) then
              TCustomForm( SourceRoot ).Designer.Modified();

          Break;
          end;

        end;

      SinkPin.Connect( NIL );
      for I := 0 to Form.Items.Count - 1 do
        begin
        SourcePin := TOWSourcePin( Form.Items.Item [ I ].SubItems.Objects[ 0 ]);
        if ( Form.Items.Item [ I ].StateIndex = 1 ) then
          begin
          SinkPin.Connect( SourcePin );
          end;

        end;

      Result := True;
      end;

  finally
  end;

end;

function SourcePinEdit( Designer : IADesigner; SourcePin : TOWSourcePin ) : Boolean;
var
  I             : Integer;
  SinkPin       : TOWSinkPin;
  SinkRoot      : TComponent;
  SinkChanged   : Boolean;

begin
  Result := False;

  if( SourcePin = NIL ) then
    Exit;

  try
    if ( Form.ExecuteForSource( Designer, SourcePin ) = mrOk ) then
      begin
      for I := 0 to Form.Items.Count - 1 do
        begin
        SinkChanged := False;
        SinkPin := TOWSinkPin( Form.Items.Item [ I ].SubItems.Objects[ 0 ]);
        if ( Form.Items.Item [ I ].StateIndex = 1 ) then
          begin
          if( not SourcePin.IsLinkedTo( SinkPin )) then
            begin
            SinkChanged := True;
            end;
          end

        else
          begin
          if( SourcePin.IsLinkedTo( SinkPin )) then
            begin
            SinkChanged := True;
            end;
          end;

        if( SinkChanged ) then
          begin
          SinkRoot := GetMainDesignOwner( SinkPin.Owner );
          if( SinkRoot is TCustomForm ) then
            if( Assigned( TCustomForm( SinkRoot ).Designer )) then
              TCustomForm( SinkRoot ).Designer.Modified();

          end;

        end;

      SourcePin.Disconnect();
      for I := 0 to Form.Items.Count - 1 do
        begin
        SinkPin := TOWSinkPin( Form.Items.Item [ I ].SubItems.Objects[ 0 ]);
        if ( Form.Items.Item [ I ].StateIndex = 1 ) then
          SourcePin.Connect( SinkPin );
          
        end;

      Result := True;
      end;

  finally
  end;

end;

type
  TOWPinListPropertyEditor = class(TPropertyEditor)
  public
    function  GetAttributes: TPropertyAttributes; override;
    function  GetValue: string; override;
    procedure GetProperties(Proc: TAGetPropProc); override;
    
  public
    procedure CheckRefresh();
  end;

  TOWPinListOwnerPropertyEditor = class(TOWPinListPropertyEditor)
  protected

  public
    function  GetAttributes: TPropertyAttributes; override;
    function  GetValue: string; override;
    procedure SetValue(const Value: string); override;
  end;

  TOWSourcePinPropertyEditor = class(TPropertyEditor)
  
  public
    function  GetPin() : TOWSourcePin;

  protected
    
  public
    function  GetAttributes: TPropertyAttributes; override;
    procedure Edit; override;
    function  GetValue: string; override;
    procedure SetValue(const Value: string); override;
  end;

{$IFDEF D6}
  TOWSourcePinListPropertyEditor = class(TBasePropertyEditor, IProperty)
  protected
    function GetEditValue(out Value: string): Boolean; virtual;
    function HasInstance(Instance: TPersistent): Boolean;

  protected
    procedure Activate; virtual;
    procedure Initialize; override;
    function  AllEqual: Boolean; virtual;
    function  AutoFill: Boolean; virtual;
    function  GetEditLimit: Integer; virtual;
    procedure GetProperties(Proc: TGetPropProc); virtual;
    function  GetPropInfo: PPropInfo; virtual;
    function  GetPropType: PTypeInfo;
    procedure SetPropEntry(Index: Integer; AInstance: TPersistent;
      APropInfo: PPropInfo); override;
    procedure Revert;
    function  ValueAvailable: Boolean;
    procedure Modified;

  public
    function  GetName: string; virtual;

  public
    function  GetAttributes: TPropertyAttributes; virtual;
    procedure GetValues(Proc: TGetStrProc); virtual;
    procedure SetValue(const Value: string); virtual;
    function  GetValue: string; virtual;
    procedure Edit; virtual;

{$ELSE}
  TOWSourcePinListPropertyEditor = class(TPropertyEditor)
  public
    function  GetName: string; override;

  public
    function  GetAttributes: TPropertyAttributes; override;
    procedure GetValues(Proc: TGetStrProc); override;
    procedure SetValue(const Value: string); override;
    function  GetValue: string; override;
    procedure Edit; override;

{$ENDIF}
  protected
    FName           : String;
    FPin            : TOWSourcePin;
    FDesigner       : IDesigner;
    FPinListEditor  : TOWPinListPropertyEditor;

  public
    constructor CreateEx(const ADesigner: IADesigner; APin : TOWSourcePin; AName : String; PinListEditor : TOWPinListPropertyEditor);
    destructor  Destroy; override;

  public
    function    GetPin() : TOWSourcePin;

  end;
  TOWSinkPinPropertyEditor = class(TPropertyEditor)
  public
    function  GetPin() : TOWSinkPin;

  public
    function  GetAttributes: TPropertyAttributes; override;
    procedure GetValues(Proc: TGetStrProc); override;
    procedure SetValue(const Value: string); override;
    function  GetValue: string; override;
    procedure Edit; override;

  end;

{$IFDEF D6}
  TOWSinkPinListPropertyEditor = class(TBasePropertyEditor, IProperty)
  protected
    function GetEditValue(out Value: string): Boolean; virtual;
    function HasInstance(Instance: TPersistent): Boolean;

  protected
    procedure Activate; virtual;
    procedure Initialize; override;
    function  AllEqual: Boolean; virtual;
    function  AutoFill: Boolean; virtual;
    procedure Edit; virtual;
    function  GetEditLimit: Integer; virtual;
    procedure GetProperties(Proc: TGetPropProc); virtual;
    function  GetPropInfo: PPropInfo; virtual;
    function  GetPropType: PTypeInfo;
    procedure SetPropEntry(Index: Integer; AInstance: TPersistent;
      APropInfo: PPropInfo); override;
    procedure Revert;
    function  ValueAvailable: Boolean;
    procedure Modified;

  public
    function  GetName: string; virtual;

  public
    function  GetAttributes: TPropertyAttributes; virtual;
    procedure GetValues(Proc: TGetStrProc); virtual;
    procedure SetValue(const Value: string); virtual;
    function  GetValue: string; virtual;

{$ELSE}
  TOWSinkPinListPropertyEditor = class(TPropertyEditor)
  public
    function  GetName: string; override;

  public
    function  GetAttributes: TPropertyAttributes; override;
    procedure GetValues(Proc: TGetStrProc); override;
    procedure SetValue(const Value: string); override;
    function  GetValue: string; override;
    procedure Edit; override;

{$ENDIF}
  public
    function  GetPin() : TOWSinkPin;
    
  protected
    FDesigner: IDesigner;

  protected
    FName    : String;
    FPin     : TOWSinkPin;
    FPinListEditor  : TOWPinListPropertyEditor;

  public
    constructor CreateEx(const ADesigner: IADesigner; APin : TOWSinkPin; AName : String; PinListEditor : TOWPinListPropertyEditor );
    destructor  Destroy; override;

  end;

function TOWSourcePinPropertyEditor.GetAttributes: TPropertyAttributes;
begin
  Result := [ paDialog ];
  RequestRefresh();
end;

procedure TOWSourcePinPropertyEditor.Edit;
var
  SourcePin             : TOWSourcePin;

begin
  RequestRefreshEx(Designer);
//  SourcePin := TOWSourcePin( GetOrdValue ());
  SourcePin := GetPin();
//  SourcePinEdit( SourcePin );
  if( SourcePinEdit( Designer, SourcePin )) then
    Modified();

end;

procedure TOWSourcePinPropertyEditor.SetValue(const Value: string);
var
  SourcePin     : TOWSourcePin;

begin
  inherited SetValue( Value );

  SourcePin := GetPin();
  if( SetSourcePinValue( NIL, SourcePin, Value )) then
    Modified();

end;

function TOWSourcePinPropertyEditor.GetValue: string;
var
  SourcePin : TOWSourcePin;

begin
//  inherited GetValue();
  RequestRefreshEx(Designer);
  SourcePin := GetPin();
  Result := GetSourcePinValue( SourcePin, Designer );

end;

//------------------------------------------------------------------------------
function  TOWSourcePinPropertyEditor.GetPin() : TOWSourcePin;
begin
  Result := TOWSourcePin( GetOrdValue());
end;
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
constructor TOWSourcePinListPropertyEditor.CreateEx(const ADesigner: IADesigner; APin : TOWSourcePin; AName : String; PinListEditor : TOWPinListPropertyEditor );
begin
  inherited Create( ADesigner, 1 );
  FDesigner := ADesigner;
  FPin  := APin;
  FName := AName;
  FPinListEditor := PinListEditor;
  FPin.CurrentEditorPtr := @FPin;
end;
//------------------------------------------------------------------------------
destructor  TOWSourcePinListPropertyEditor.Destroy;
begin
  if( FPin <> NIL )then
    FPin.CurrentEditorPtr := NIL;
    
  inherited;
end;
//------------------------------------------------------------------------------
function  TOWSourcePinListPropertyEditor.GetPin() : TOWSourcePin;
begin
  Result := FPin;
end;
//---------------------------------------------------------------------------
function  TOWSourcePinListPropertyEditor.GetName(): string;
begin
  Result := FName;
end;
//---------------------------------------------------------------------------
function  TOWSourcePinListPropertyEditor.GetAttributes: TPropertyAttributes;
begin
  Result := [ paDialog ];
  RequestRefresh();
end;
//---------------------------------------------------------------------------
procedure TOWSourcePinListPropertyEditor.GetValues(Proc: TGetStrProc);
begin
end;
//---------------------------------------------------------------------------
procedure TOWSourcePinListPropertyEditor.SetValue(const Value: string);
var
  SourcePin     : TOWSourcePin;
  
begin
  SourcePin := GetPin();
  if( SetSourcePinValue( NIL, SourcePin, Value )) then
    Modified();

end;
//---------------------------------------------------------------------------
function  TOWSourcePinListPropertyEditor.GetValue: string;
var
  SourcePin : TOWSourcePin;

begin
  RequestRefreshEx(Designer);
  SourcePin := GetPin();
  Result := GetSourcePinValue( SourcePin, Designer );
  
  FPinListEditor.CheckRefresh();
end;
//---------------------------------------------------------------------------
{$IFDEF D6}
function TOWSourcePinListPropertyEditor.GetEditValue(out Value: string): Boolean;
begin
  Value := GetValue();
  Result := True;
end;

function TOWSourcePinListPropertyEditor.HasInstance(Instance: TPersistent): Boolean;
begin
  Result := True;
end;

procedure TOWSourcePinListPropertyEditor.Activate;
begin
end;

procedure TOWSourcePinListPropertyEditor.Initialize;
begin
end;

function TOWSourcePinListPropertyEditor.AllEqual: Boolean;
begin
  Result := True;
end;

function TOWSourcePinListPropertyEditor.AutoFill: Boolean;
begin
  Result := False;
end;

function TOWSourcePinListPropertyEditor.GetEditLimit: Integer;
begin
  Result := 255;
end;

procedure TOWSourcePinListPropertyEditor.GetProperties(Proc: TGetPropProc);
begin
end;

function TOWSourcePinListPropertyEditor.GetPropInfo: PPropInfo;
begin
  Result := NIL;
end;

function TOWSourcePinListPropertyEditor.GetPropType: PTypeInfo;
begin
  Result := typeinfo( TOWSourcePin );
end;

procedure TOWSourcePinListPropertyEditor.SetPropEntry(Index: Integer; AInstance: TPersistent;
      APropInfo: PPropInfo);
begin
end;

procedure TOWSourcePinListPropertyEditor.Revert;
begin
end;

function TOWSourcePinListPropertyEditor.ValueAvailable: Boolean;
begin
  Result := True;
end;

procedure TOWSourcePinListPropertyEditor.Modified;
begin
  if FDesigner <> nil then
    FDesigner.Modified;
end;

{$ENDIF}
//---------------------------------------------------------------------------
procedure TOWSourcePinListPropertyEditor.Edit;
var
  SourcePin     : TOWSourcePin;

begin
  SourcePin := GetPin();
  if( SourcePinEdit( Designer, SourcePin )) then
      Modified();
end;
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure TOWSinkPinPropertyEditor.Edit;
var
  SinkPin : TOWSinkPin;

begin
  RequestRefreshEx(Designer);
  SinkPin := GetPin();
  if( SinkPinEdit( Designer, SinkPin )) then
    Modified();

end;
//------------------------------------------------------------------------------
function  TOWSinkPinPropertyEditor.GetPin() : TOWSinkPin;
begin
  Result := TOWSinkPin( GetOrdValue());
end;
//------------------------------------------------------------------------------
function TOWSinkPinPropertyEditor.GetAttributes: TPropertyAttributes;
begin
  Result := [ paValueList, paSortList, paDialog ];
  RequestRefresh();
end;
//------------------------------------------------------------------------------
procedure TOWSinkPinPropertyEditor.GetValues(Proc: TGetStrProc);
var
  SinkPin       : TOWSinkPin;

begin
  SinkPin := GetPin();
  GetSinkPinValues( SinkPin, Proc );
end;
//------------------------------------------------------------------------------
procedure TOWSinkPinPropertyEditor.SetValue(const Value: string);
var
  SinkPin       : TOWSinkPin;

begin
  SinkPin := GetPin();
  if( SetSinkPinValue( NIL, SinkPin, Value )) then
    Modified();

//  inherited SetValue( Value );
end;
//------------------------------------------------------------------------------
function  TOWSinkPinPropertyEditor.GetValue: string;
var
  SinkPin   : TOWSinkPin;

begin
//  inherited GetValue();
  RequestRefreshEx(Designer);
  SinkPin := GetPin();
  Result := GetSinkPinValue( SinkPin, Designer );
end;

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//------------------------------------------------------------------------------
constructor TOWSinkPinListPropertyEditor.CreateEx(const ADesigner: IADesigner; APin : TOWSinkPin; AName : String; PinListEditor : TOWPinListPropertyEditor );
begin
  inherited Create( ADesigner, 1 );
  FDesigner := ADesigner; 
  FPin  := APin;
  FName := AName;
  FPinListEditor := PinListEditor;
  FPin.CurrentEditorPtr := @FPin;
end;
//---------------------------------------------------------------------------
destructor  TOWSinkPinListPropertyEditor.Destroy;
begin
  if( FPin <> NIL )then
    FPin.CurrentEditorPtr := NIL;

  inherited;
end;
//------------------------------------------------------------------------------
{$IFDEF D6}
function TOWSinkPinListPropertyEditor.GetEditValue(out Value: string): Boolean;
begin
  Value := GetValue();
  Result := True;
  FPinListEditor.CheckRefresh();
end;

function TOWSinkPinListPropertyEditor.HasInstance(Instance: TPersistent): Boolean;
begin
  Result := True;
end;

procedure TOWSinkPinListPropertyEditor.Activate;
begin
end;

procedure TOWSinkPinListPropertyEditor.Initialize;
begin
end;

function TOWSinkPinListPropertyEditor.AllEqual: Boolean;
begin
  Result := True;
end;

function TOWSinkPinListPropertyEditor.AutoFill: Boolean;
begin
  Result := False;
end;

function TOWSinkPinListPropertyEditor.GetEditLimit: Integer;
begin
  Result := 255;
end;

procedure TOWSinkPinListPropertyEditor.GetProperties(Proc: TGetPropProc);
begin
end;

function TOWSinkPinListPropertyEditor.GetPropInfo: PPropInfo;
begin
//  Result := FPropList^[0].PropInfo;
  Result := NIL;
end;

function TOWSinkPinListPropertyEditor.GetPropType: PTypeInfo;
begin
  Result := typeinfo( TOWSinkPin );
end;

procedure TOWSinkPinListPropertyEditor.SetPropEntry(Index: Integer; AInstance: TPersistent;
      APropInfo: PPropInfo);
begin
end;

procedure TOWSinkPinListPropertyEditor.Revert;
begin
end;

function TOWSinkPinListPropertyEditor.ValueAvailable: Boolean;
begin
  Result := True;
end;

procedure TOWSinkPinListPropertyEditor.Modified;
begin
  if FDesigner <> nil then
    FDesigner.Modified;
end;

{$ENDIF}
//------------------------------------------------------------------------------
procedure TOWSinkPinListPropertyEditor.Edit;
var
  SinkPin : TOWSinkPin;

begin
  RequestRefreshEx(Designer);
  SinkPin := GetPin();
  if( SinkPinEdit( Designer, SinkPin )) then
    Modified();

end;
//------------------------------------------------------------------------------
function  TOWSinkPinListPropertyEditor.GetPin() : TOWSinkPin;
begin
  Result := FPin;
end;
//---------------------------------------------------------------------------
function  TOWSinkPinListPropertyEditor.GetName: string;
begin
  Result := FName;
end;
//---------------------------------------------------------------------------
function TOWSinkPinListPropertyEditor.GetAttributes: TPropertyAttributes;
begin
  Result := [ paValueList, paSortList, paDialog ];
  RequestRefresh();
end;
//---------------------------------------------------------------------------
procedure TOWSinkPinListPropertyEditor.GetValues(Proc: TGetStrProc);
var
  SinkPin       : TOWSinkPin;

begin
  SinkPin := GetPin();
  GetSinkPinValues( SinkPin, Proc );
end;
//---------------------------------------------------------------------------
procedure TOWSinkPinListPropertyEditor.SetValue(const Value: string);
var
  SinkPin       : TOWSinkPin;

begin
  SinkPin := GetPin();
  if( SetSinkPinValue( NIL, SinkPin, Value )) then
    Modified();

//  inherited SetValue( Value );
end;
//---------------------------------------------------------------------------
function  TOWSinkPinListPropertyEditor.GetValue: string;
var
  SinkPin   : TOWSinkPin;

begin
//  inherited GetValue();
  RequestRefreshEx(Designer);
  SinkPin := GetPin();
  Result := GetSinkPinValue( SinkPin, Designer );
  FPinListEditor.CheckRefresh();
end;

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//------------------------------------------------------------------------------
type TOWInputPinsCategory = class( TPropertyCategory )
  class function Name: String; override;
  class function Description: String; override;
end;
//------------------------------------------------------------------------------
class function TOWInputPinsCategory.Name: String;
begin
  Result := 'Input Pins';
end;
//------------------------------------------------------------------------------
class function TOWInputPinsCategory.Description: String;
begin
  Result := 'OpenWire Input Pins';
end;
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
type TOWOutputPinsCategory = class( TPropertyCategory )
  class function Name: String; override;
  class function Description: String; override;
end;
//------------------------------------------------------------------------------
class function TOWOutputPinsCategory.Name: String;
begin
  Result := 'Output Pins';
end;
//------------------------------------------------------------------------------
class function TOWOutputPinsCategory.Description: String;
begin
  Result := 'OpenWire Output Pins';
end;
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
type TOWPinListCategory = class( TPropertyCategory )
  class function Name: String; override;
  class function Description: String; override;
end;
//------------------------------------------------------------------------------
class function TOWPinListCategory.Name: String;
begin
  Result := 'Pin Lists';
end;
//------------------------------------------------------------------------------
class function TOWPinListCategory.Description: String;
begin
  Result := 'OpenWire Lists of Pins';
end;
//------------------------------------------------------------------------------
procedure Register;
begin
  RegisterPropertyInCategory( TOWInputPinsCategory,     typeinfo(TOWSinkPin) ); 
  RegisterPropertyInCategory( TOWOutputPinsCategory,    typeinfo(TOWSourcePin) );
  RegisterPropertyInCategory( TOWPinListCategory,       typeinfo(TOWPinList) );

  RegisterPropertyEditor( typeinfo(TOWSourcePin),       NIL, '', TOWSourcePinPropertyEditor);
  RegisterPropertyEditor( typeinfo(TOWSinkPin),         NIL, '', TOWSinkPinPropertyEditor);
  RegisterPropertyEditor( typeinfo(TOWPinList),         NIL, '', TOWPinListPropertyEditor);
  RegisterPropertyEditor( typeinfo(TOWPinListOwner),    NIL, '', TOWPinListOwnerPropertyEditor);
end;

{$R *.DFM}

procedure TSourcePinForm.LinkAllButtonClick(Sender: TObject);
var
  I : Integer;
begin
  for I := 0 to ListView.Items.Count - 1 do
    ListView.Items.Item [ I ].Checked := True;
end;

procedure TSourcePinForm.UnlinkAllButtonClick(Sender: TObject);
var
  I : Integer;
begin
  for I := 0 to ListView.Items.Count - 1 do
    ListView.Items.Item [ I ].Checked := False;
end;

procedure TSourcePinForm.RestoreButtonClick(Sender: TObject);
var
  I : Integer;
begin
  for I := 0 to ListView.Items.Count - 1 do
    begin
    if ( TImageIndex(ListView.Items.Item[ I ].Data) = -1 ) then
      ListView.Items.Item[ I ].StateIndex := TImageIndex( ListView.Items.Item[ I ].Data )

    else
      ListView.Items.Item[ I ].Checked := Boolean( ListView.Items.Item[ I ].Data );
      
    end;
end;

procedure TSourcePinForm.FormShow(Sender: TObject);
begin
  FillFormsInfo();
//  PopulateForm( Designer.GetRoot() );
  UpdateLinksCount();
end;

procedure TSourcePinForm.FillFormsInfo();
var
  FormNames     : TOWModulesColection;
  I             : Integer;

begin
  FormNames := TOWModulesColection.Create;

  Designer.GetProjectModules( FormNames.GetModules );

  FormsComboBox.Items.Clear();
//  FormsComboBox.Items.Assign( FormNames );
  for I := 0 to FormNames.Count - 1 do
    begin
    if( CanAccessRootFromName( Designer, FormNames.Strings[ I ] ) ) then
      begin
      if( FormNames.Strings[ I ] = Root.Name ) then
        begin
        FormsComboBox.Items.Add( FormNames.Strings[ I ] + '  (Current)' );
        FormsComboBox.ItemIndex := FormsComboBox.Items.Count - 1;
        end

      else
        FormsComboBox.Items.Add( FormNames.Strings[ I ] );
      end;
    end;

  FormsComboBox.Items.Add( 'All forms' );

  PopulateAllEntries();
  if( AllSelected ) then
    FormsComboBox.ItemIndex := FormsComboBox.Items.Count - 1;

  FormsComboBoxChange( Self );

  FormNames.Free;

end;

procedure TSourcePinForm.PopulateAllEntries();
var
  I             : Integer;             
  CurrentRoot   : TComponent;
  
begin
  PinsList.Clear();  
  for I := 0 to FormsComboBox.Items.Count - 2 do
    begin
    CurrentRoot := RootFromName( FormsComboBox.Items.Strings[ I ] );
    PopulateSingleFormEntries( NIL, CurrentRoot, True );
    end;

  RootFromName( Root.Name );
end;

procedure TSourcePinForm.PopulateAll();
var
  I             : Integer;             
  CurrentRoot   : TComponent;
  
begin
  ClearData();
  for I := 0 to FormsComboBox.Items.Count - 2 do
    begin
    CurrentRoot := RootFromName( FormsComboBox.Items.Strings[ I ] );
    PopulateSingleForm( NIL, CurrentRoot, False, True );
    end;

  RootFromName( Root.Name );
end;

procedure TSourcePinForm.PopulateForm( ARootComponent : TComponent );
var
  I             : Integer;
  CurrentRoot   : TComponent;
//  MyRoot        : Boolean;

begin
  ClearData();
//  MyRoot := GetMainOwnerComponent( SourcePin.Owner ) = ARootComponent;
//  PopulateSingleForm( Root, False, False );
  PopulateSingleForm( ARootComponent, ARootComponent, False, False );

  for I := 0 to FormsComboBox.Items.Count - 2 do
    begin
    CurrentRoot := RootFromName( FormsComboBox.Items.Strings[ I ] );
    if( ARootComponent <> CurrentRoot ) then
      PopulateSingleForm( ARootComponent, CurrentRoot, True, True );

    end;

end;

function TSourcePinForm.LinkToDisplayString( ARootComponent : TComponent; LinkStr : String; FullPath : Boolean; OwnRoot : Boolean; var PinStr : String ) : String;
var
  StrFull : String;
  Str     : String;
  iPos    : Integer;
  
begin
  if( not OwnRoot ) then
    begin
    iPos := Pos( '.', LinkStr );
    StrFull := LinkStr;
    Delete( StrFull, 1, iPos );
    end

  else
    StrFull := LinkStr;
    
  iPos := Pos( '.', StrFull );
  Str := StrFull;
  Delete( Str, 1, iPos );
  PinStr := Str;
  Str := StrFull;
  iPos := Pos( '.', Str );
  Delete( Str, iPos, 10000 );
  if( FullPath ) then
    Str := ARootComponent.Name + '.' + Str;

  Result := Str;
end;

procedure TSourcePinForm.PopulateSingleFormEntries( CurrentRoot : TComponent; ARootComponent : TComponent; FullPath : Boolean );
begin
  if( SourcePin <> NIL ) then
    PopulateSingleSourceFormEntries( CurrentRoot, ARootComponent, FullPath )

  else
    PopulateSingleSinkFormEntries( CurrentRoot, ARootComponent, FullPath );

end;

procedure TSourcePinForm.PopulateSingleForm( CurrentRoot : TComponent; ARootComponent : TComponent; OnlyConnected : Boolean; FullPath : Boolean );
begin
  ListUpdating := True;
  if( SourcePin <> NIL ) then
    PopulateSingleSourceForm( CurrentRoot, ARootComponent, OnlyConnected, FullPath )

  else
    PopulateSingleSinkForm( CurrentRoot, ARootComponent, OnlyConnected, FullPath );

  ListUpdating := False;
end;

procedure TSourcePinForm.PopulateSingleSinkFormEntries( CurrentRoot : TComponent; ARootComponent : TComponent; FullPath : Boolean );
var
  Values        : TStringList;
  I             : Integer;
  SourcePin     : TOWSourcePin;
//  Item          : TListItem;
  Entry         : TOWPinEntry;
  iPos          : Integer;
  Str           : String;
//  StrFull       : String;
  LinkedTo      : String;
  LinkType      : String;
//  Name          : String;
  OwnRoot       : Boolean;
//  SourceRoot    : TComponent;
  OwnDataModule : Boolean;
  LinkName      : String;
  PinStr        : String;

  FullSinkPath  : Boolean;

begin
  Items.BeginUpdate();
  Values := TStringList.Create;
  try
    OwnRoot := ( ARootComponent = OWGetMainOwnerComponent( SinkPin.Owner ));
    OwnDataModule := False;

    if( not OwnRoot ) then
      OwnDataModule := ( ARootComponent.Name = OWGetMainOwnerComponent( SinkPin.Owner ).Name );

    GetPinValueList( ARootComponent, SinkPin, Values );
    for I := 0 to Values.Count - 1 do
      begin
      SourcePin := TOWSourcePin( Values.Objects[ I ] );
//      if( OnlyConnected ) then
//        if( not SourcePin.IsLinkedTo( SinkPin ) ) then
//        if( SinkPin.GetLinkStr(False) <> Values.Strings[ I ] ) then
//          if( TOWAccessSourcePin( SourcePin ).FLinkStorage = NIL ) then
//          Continue;

      Entry := TOWPinEntry.Create();
      PinsList.Add( Entry );

//      Item := Items.Add ();
//      Item.ImageIndex := -1;

      Str := LinkToDisplayString( ARootComponent, Values.Strings [ I ], FullPath, OwnRoot, PinStr );
      Entry.PinName := PinStr;
//      Item.Caption := PinStr;

//      Item.SubItems.AddObject ( Str, Values.Objects [ I ] );
      Entry.OwnerName := Str; 
      Entry.Pin := SourcePin;
      Entry.CanConnect := True;
      if ( not SinkPin.CanConnectTo( SourcePin )) then
        begin
        Entry.SavedChecked := False;
        Entry.Checked := False;
        Entry.CanConnect := False;
//        Item.StateIndex := -1;
//        Item.Data := Pointer( -1 );
        end

      else
        begin
//        Item.Checked := SourcePin.IsLinkedTo ( SinkPin );
        LinkName := Values.Strings[ I ];
        if( OwnDataModule ) then
          begin
          iPos := Pos( '.', LinkName );
          Delete( LinkName, 1, iPos );
          end;

//        Item.Checked := ( SinkPin.GetLinkStr(False) = LinkName );
//        Item.Data := Pointer( Item.Checked );
//        Item.StateIndex := Ord( Item.Checked );
        Entry.SavedChecked := ( SinkPin.GetLinkStr(False) = LinkName );
        Entry.Checked := Entry.SavedChecked;

//        Str := LinkToDisplayString( ARootComponent, Values.Strings [ I ], FullPath, OwnRoot, PinStr );
//        LinkName := Str + '.' + PinStr;
        
        end;

      LinkedTo := '';
      LinkType := '';
      if ( SourcePin.IsConnected() ) then
        begin
        LinkedTo := GetSourcePinValue( SourcePin, Designer );
        FullSinkPath := True;
        if( SourcePin.GetSinkCountAll() = 1 ) then
          begin
          if( SourcePin.SinkCount > 0 ) then
            begin
            OwnRoot := ( OWGetMainOwnerComponent( SourcePin.Sinks[ 0 ].Owner ) = OWGetMainOwnerComponent( SourcePin.Owner ));
            FullSinkPath := ( OWGetMainOwnerComponent( SourcePin.Sinks[ 0 ].Owner ) <> CurrentRoot );
            end;

          Str := LinkToDisplayString( OWGetMainOwnerComponent( SourcePin.Sinks[ 0 ].Owner ), LinkedTo, FullSinkPath, OwnRoot, PinStr );
          LinkedTo := Str + '.' + PinStr;
          end;

        if( SourcePin.Sinks[ 0 ].DownStreamLinkName <> '' ) then
          LinkType := 'Downstream(' + SourcePin.Sinks[0].DownStreamLinkName + ')    ';

        if( SourcePin.Sinks[ 0 ].UpStreamLinkName <> '' ) then
          LinkType := LinkType + 'Upstream(' + SourcePin.Sinks[0].UpStreamLinkName + ')';

        end;

      Entry.LinkedTo := LinkedTo;
      Entry.LinkType := LinkType;
//      Item.SubItems.Add ( LinkedTo );
//      Item.SubItems.Add ( LinkType );
//      Item.SubItems.Add ( Values.Strings[ I ] );
      end;
      
  finally
    Values.Free;
    Items.EndUpdate();

  end;

end;

procedure TSourcePinForm.PopulateSingleSourceFormEntries( CurrentRoot : TComponent; ARootComponent : TComponent; FullPath : Boolean );
var
  Values        : TStringList;
  I             : Integer;
  SinkPin       : TOWSinkPin;
//  Item          : TListItem;
  Entry         : TOWPinEntry;
  iPos          : Integer;
  Str           : String;
//  StrFull       : String;
  LinkedTo      : String;
  LinkType      : String;
//  Name          : String;
  OwnRoot       : Boolean;
  OwnDataModule : Boolean;
  LinkName      : String;
  PinStr        : String;

  FullSinkPath  : Boolean;

begin
  Items.BeginUpdate();
  Values := TStringList.Create;
  try
    OwnRoot := ( ARootComponent = OWGetMainOwnerComponent( SourcePin.Owner ));
    OwnDataModule := False;

    if( not OwnRoot ) then
      OwnDataModule := ( ARootComponent.Name = OWGetMainOwnerComponent( SourcePin.Owner ).Name );

    GetPinValueList( ARootComponent, SourcePin, Values );
    for I := 0 to Values.Count - 1 do
      begin
      SinkPin := TOWSinkPin( Values.Objects[ I ] );
//      if( OnlyConnected ) then
//        if( not SourcePin.IsLinkedTo( SinkPin ) ) then
//        if( not SourcePin.IsLinkedTo( Values.Strings[ I ] ) ) then
//          if( TOWAccessSourcePin( SourcePin ).FLinkStorage = NIL ) then
//          Continue;

//      Item := Items.Add ();
//      Item.ImageIndex := -1;
      Entry := TOWPinEntry.Create();
      PinsList.Add( Entry );

      Str := LinkToDisplayString( ARootComponent, Values.Strings [ I ], FullPath, OwnRoot, PinStr );
      Entry.PinName := PinStr;
//      Item.Caption := PinStr;
      
//      Item.SubItems.AddObject ( Str, Values.Objects [ I ] );
      Entry.OwnerName := Str;
      Entry.Pin := SinkPin;
      Entry.CanConnect := True;
      
      if ( not SourcePin.CanConnectTo( SinkPin )) then
        begin
        Entry.Checked := False;
        Entry.SavedChecked := False;
        Entry.CanConnect := False;
//        Item.StateIndex := -1;
//        Item.Data := Pointer( -1 );
        end

      else
        begin
//        Item.Checked := SourcePin.IsLinkedTo ( SinkPin );
        LinkName := Values.Strings[ I ];
        if( OwnDataModule ) then
          begin
          iPos := Pos( '.', LinkName );
          Delete( LinkName, 1, iPos );
          end;

//        Item.Checked := SourcePin.IsLinkedTo ( LinkName );
//        Item.Data := Pointer( Item.Checked );
//        Item.StateIndex := Ord( Item.Checked );
        Entry.Checked := SourcePin.IsLinkedTo ( LinkName );
        Entry.SavedChecked := Entry.Checked;

//        Str := LinkToDisplayString( ARootComponent, Values.Strings [ I ], FullPath, OwnRoot, PinStr );
//        LinkName := Str + '.' + PinStr;
        
        end;

      LinkedTo := '';
      LinkType := '';
      if ( SinkPin.SourcePin <> NIL ) then
        begin
        LinkedTo := SinkPin.GetLinkStr( False );
        OwnRoot := ( OWGetMainOwnerComponent( SinkPin.SourcePin.Owner ) = OWGetMainOwnerComponent( SinkPin.Owner ));
        FullSinkPath := ( OWGetMainOwnerComponent( SinkPin.SourcePin.Owner ) <> CurrentRoot );
{
          if( SourcePin.SinkCount > 0 ) then
            begin
            OwnRoot := ( GetMainOwnerComponent( SourcePin.Sinks[ 0 ].Owner ) = GetMainOwnerComponent( SourcePin.Owner ));
            FullSinkPath := ( GetMainOwnerComponent( SourcePin.Sinks[ 0 ].Owner ) <> CurrentRoot );
            end;

          Str := LinkToDisplayString( GetMainOwnerComponent( SourcePin.Sinks[ 0 ].Owner ), LinkedTo, FullSinkPath, OwnRoot, PinStr );
}
//        Str := LinkToDisplayString( ARootComponent, LinkedTo, FullPath, OwnRoot, PinStr );
        Str := LinkToDisplayString( OWGetMainOwnerComponent( SinkPin.SourcePin.Owner ), LinkedTo, FullSinkPath, OwnRoot, PinStr );
        LinkedTo := Str + '.' + PinStr;
//        LinkedTo := ValueToString ( SinkPin.SourcePin, '.' );

        if( SinkPin.DownStreamLinkName <> '' ) then
          LinkType := 'Downstream(' + SinkPin.DownStreamLinkName + ')    ';

        if( SinkPin.UpStreamLinkName <> '' ) then
          LinkType := LinkType + 'Upstream(' + SinkPin.UpStreamLinkName + ')';

        end;

      Entry.LinkedTo := LinkedTo;
      Entry.LinkType := LinkType;
//      Item.SubItems.Add ( LinkedTo );
//      Item.SubItems.Add ( LinkType );
//      Item.SubItems.Add ( Values.Strings[ I ] );
      end;

  finally
    Values.Free;
    Items.EndUpdate();

  end;

end;

function TSourcePinForm.EntryFromPin ( Pin : TOWPin ) : TOWPinEntry;
var
  I : Integer;
  
begin
  Result := NIL;
  for I := 0 to PinsList.Count - 1 do
    if( TOWPinEntry( PinsList.Items[ I ] ).Pin = Pin ) then
      begin
      Result := TOWPinEntry( PinsList.Items[ I ] );
      Break;
      end;
    

end;

procedure TSourcePinForm.PopulateSingleSinkForm( CurrentRoot : TComponent; ARootComponent : TComponent; OnlyConnected : Boolean; FullPath : Boolean );
var
  Values        : TStringList;
  I             : Integer;
  SourcePin     : TOWSourcePin;
  Item          : TListItem;
  iPos          : Integer;
  Str           : String;
//  StrFull       : String;
  LinkedTo      : String;
  LinkType      : String;
//  Name          : String;
  OwnRoot       : Boolean;
//  SourceRoot    : TComponent;
  OwnDataModule : Boolean;
  LinkName      : String;
  PinStr        : String;
  
  FullSinkPath  : Boolean;
  
  Entry         : TOWPinEntry;
  Linked        : Boolean;

begin
  Items.BeginUpdate();
  Values := TStringList.Create;
  try
    OwnRoot := ( ARootComponent = OWGetMainOwnerComponent( SinkPin.Owner ));
    OwnDataModule := False;

    if( not OwnRoot ) then
      OwnDataModule := ( ARootComponent.Name = OWGetMainOwnerComponent( SinkPin.Owner ).Name );

    GetPinValueList( ARootComponent, SinkPin, Values );
    for I := 0 to Values.Count - 1 do
      begin
      SourcePin := TOWSourcePin( Values.Objects[ I ] );
      Entry := EntryFromPin( SourcePin );
      if( OnlyConnected ) then
        begin
        if( Entry <> NIL ) then
          begin
          if( not Entry.Checked ) then
            Continue;
        
          end

        else
          if( SinkPin.GetLinkStr(False) <> Values.Strings[ I ] ) then
            Continue;
            
        end;

//        if( not SourcePin.IsLinkedTo( SinkPin ) ) then
//        if( SinkPin.GetLinkStr(False) <> Values.Strings[ I ] ) then
//          if( TOWAccessSourcePin( SourcePin ).FLinkStorage = NIL ) then

      Item := Items.Add ();
      Item.ImageIndex := -1;

      Str := LinkToDisplayString( ARootComponent, Values.Strings [ I ], FullPath, OwnRoot, PinStr );
      Item.Caption := PinStr;

      Item.SubItems.AddObject ( Str, Values.Objects [ I ] );
      if ( not SinkPin.CanConnectTo( SourcePin )) then
        begin
        Item.StateIndex := -1;
        Item.Data := Pointer( -1 );
        end

      else
        begin
//        Item.Checked := SourcePin.IsLinkedTo ( SinkPin );
        LinkName := Values.Strings[ I ];
        if( OwnDataModule ) then
          begin
          iPos := Pos( '.', LinkName );
          Delete( LinkName, 1, iPos );
          end;

        Linked := ( SinkPin.GetLinkStr(False) = LinkName );

        if( Entry <> NIL ) then
          Item.Checked := Entry.Checked

        else
          Item.Checked := Linked;

        Item.Data := Pointer( Linked );
        Item.StateIndex := Ord( Item.Checked );

//        Str := LinkToDisplayString( ARootComponent, Values.Strings [ I ], FullPath, OwnRoot, PinStr );
//        LinkName := Str + '.' + PinStr;

        end;

      LinkedTo := '';
      LinkType := '';
      if ( SourcePin.IsConnected() ) then
        begin
        LinkedTo := GetSourcePinValue( SourcePin, Designer );
        FullSinkPath := True;
        if( SourcePin.GetSinkCountAll() = 1 ) then
          begin
          if( SourcePin.SinkCount > 0 ) then
            begin
            OwnRoot := ( OWGetMainOwnerComponent( SourcePin.Sinks[ 0 ].Owner ) = OWGetMainOwnerComponent( SourcePin.Owner ));
            FullSinkPath := ( OWGetMainOwnerComponent( SourcePin.Sinks[ 0 ].Owner ) <> CurrentRoot );
            end;

          Str := LinkToDisplayString( OWGetMainOwnerComponent( SourcePin.Sinks[ 0 ].Owner ), LinkedTo, FullSinkPath, OwnRoot, PinStr );
          LinkedTo := Str + '.' + PinStr;
          end;

        if( SourcePin.Sinks[ 0 ].DownStreamLinkName <> '' ) then
          LinkType := 'Downstream(' + SourcePin.Sinks[0].DownStreamLinkName + ')    ';

        if( SourcePin.Sinks[ 0 ].UpStreamLinkName <> '' ) then
          LinkType := LinkType + 'Upstream(' + SourcePin.Sinks[0].UpStreamLinkName + ')';

        end;

      Item.SubItems.AddObject ( LinkedTo, Entry );
      Item.SubItems.Add ( LinkType );
      Item.SubItems.Add ( Values.Strings[ I ] );
      end;
      
  finally
    Values.Free;
    Items.EndUpdate();

  end;

end;

procedure TSourcePinForm.PopulateSingleSourceForm( CurrentRoot : TComponent; ARootComponent : TComponent; OnlyConnected : Boolean; FullPath : Boolean );
var
  Values        : TStringList;
  I             : Integer;
  SinkPin       : TOWSinkPin;
  Item          : TListItem;
  iPos          : Integer;
  Str           : String;
//  StrFull       : String;
  LinkedTo      : String;
  LinkType      : String;
//  Name          : String;
  OwnRoot       : Boolean;
  OwnDataModule : Boolean;
  LinkName      : String;
  PinStr        : String;

  FullSinkPath  : Boolean;
  Entry         : TOWPinEntry;

  Linked        : Boolean;

begin
  Items.BeginUpdate();
  Values := TStringList.Create;
  try
    OwnRoot := ( ARootComponent = OWGetMainOwnerComponent( SourcePin.Owner ));
    OwnDataModule := False;

    if( not OwnRoot ) then
      OwnDataModule := ( ARootComponent.Name = OWGetMainOwnerComponent( SourcePin.Owner ).Name );

    GetPinValueList( ARootComponent, SourcePin, Values );
    for I := 0 to Values.Count - 1 do
      begin
      SinkPin := TOWSinkPin( Values.Objects[ I ] );
      Entry := EntryFromPin( SinkPin );
      if( OnlyConnected ) then
        begin
        if( Entry <> NIL ) then
          begin
          if( not Entry.Checked ) then
            Continue;
            
          end
          
        else
          if( not SourcePin.IsLinkedTo( Values.Strings[ I ] ) ) then
            Continue;
            
        end;
            
//        if( not SourcePin.IsLinkedTo( SinkPin ) ) then
//        if( not SourcePin.IsLinkedTo( Values.Strings[ I ] ) ) then
//          if( TOWAccessSourcePin( SourcePin ).FLinkStorage = NIL ) then

      Item := Items.Add ();
      Item.ImageIndex := -1;

      Str := LinkToDisplayString( ARootComponent, Values.Strings [ I ], FullPath, OwnRoot, PinStr );
      Item.Caption := PinStr;

      Item.SubItems.AddObject ( Str, Values.Objects [ I ] );
      if ( not SourcePin.CanConnectTo( SinkPin )) then
        begin
        Item.StateIndex := -1;
        Item.Data := Pointer( -1 );
        end

      else
        begin
//        Item.Checked := SourcePin.IsLinkedTo ( SinkPin );
        LinkName := Values.Strings[ I ];
        if( OwnDataModule ) then
          begin
          iPos := Pos( '.', LinkName );
          Delete( LinkName, 1, iPos );
          end;

        Linked := SourcePin.IsLinkedTo ( LinkName );

        if( Entry <> NIL ) then
          Item.Checked := Entry.Checked
           
        else
          Item.Checked := Linked;
          
        Item.Data := Pointer( Linked );
        Item.StateIndex := Ord( Item.Checked );
        
//        Str := LinkToDisplayString( ARootComponent, Values.Strings [ I ], FullPath, OwnRoot, PinStr );
//        LinkName := Str + '.' + PinStr;
        
        end;

      LinkedTo := '';
      LinkType := '';
      if ( SinkPin.SourcePin <> NIL ) then
        begin
        LinkedTo := SinkPin.GetLinkStr( False );
        OwnRoot := ( OWGetMainOwnerComponent( SinkPin.SourcePin.Owner ) = OWGetMainOwnerComponent( SinkPin.Owner ));
        FullSinkPath := ( OWGetMainOwnerComponent( SinkPin.SourcePin.Owner ) <> CurrentRoot );
{
          if( SourcePin.SinkCount > 0 ) then
            begin
            OwnRoot := ( GetMainOwnerComponent( SourcePin.Sinks[ 0 ].Owner ) = GetMainOwnerComponent( SourcePin.Owner ));
            FullSinkPath := ( GetMainOwnerComponent( SourcePin.Sinks[ 0 ].Owner ) <> CurrentRoot );
            end;

          Str := LinkToDisplayString( GetMainOwnerComponent( SourcePin.Sinks[ 0 ].Owner ), LinkedTo, FullSinkPath, OwnRoot, PinStr );
}
//        Str := LinkToDisplayString( ARootComponent, LinkedTo, FullPath, OwnRoot, PinStr );
        Str := LinkToDisplayString( OWGetMainOwnerComponent( SinkPin.SourcePin.Owner ), LinkedTo, FullSinkPath, OwnRoot, PinStr );
        LinkedTo := Str + '.' + PinStr;
//        LinkedTo := ValueToString ( SinkPin.SourcePin, '.' );

        if( SinkPin.DownStreamLinkName <> '' ) then
          LinkType := 'Downstream(' + SinkPin.DownStreamLinkName + ')    ';

        if( SinkPin.UpStreamLinkName <> '' ) then
          LinkType := LinkType + 'Upstream(' + SinkPin.UpStreamLinkName + ')';

        end;

      Item.SubItems.AddObject ( LinkedTo, Entry );
      Item.SubItems.Add ( LinkType );
      Item.SubItems.Add ( Values.Strings[ I ] );
      end;

  finally
    Values.Free;
    Items.EndUpdate();

  end;

end;

procedure TSourcePinForm.OWMUpdateInspector(var Message: TMessage);
var
  List          : TADesignerSelectionList;
  Designer      : IADesigner;

begin
    InRefresh := False;

    Designer := IADesigner( Message.WParam );
    List := TADesignerSelectionList.Create;
    Designer.GetSelections( List );
    Designer.SetSelections( EmptyList );
    Designer.SetSelections( List );
    List.Free();
end;

procedure TSourcePinForm.OWMUpdate(var Message: TMessage);
begin
    LinkAwaitsLinkingAllForms();
end;

function  TSourcePinForm.ExecuteForSource( ADesigner : IADesigner; ASourcePin : TOWSourcePin ): Integer;
begin
  Designer := ADesigner;
  SourcePin := ASourcePin;
  SinkPin := NIL;
  Root := ADesigner.GetRoot();

  LinkAllButton.Visible := True;
  UnlinkAllButton.Visible := True;

  ListView.Columns.Items[0].Caption := 'Sink pin';
  ListView.Columns.Items[2].Caption := 'Linked to';

  LinkAwaitsLinkingAllForms();
  Result := ShowModal();
end;

function  TSourcePinForm.ExecuteForSink( ADesigner : IADesigner; ASinkPin : TOWSinkPin ): Integer;
begin
  Designer := ADesigner;
  SinkPin := ASinkPin;
  SourcePin := NIL;
  Root := ADesigner.GetRoot();
  
  LinkAllButton.Visible := False;
  UnlinkAllButton.Visible := False;
  
  ListView.Columns.Items[0].Caption := 'Source pin';
  ListView.Columns.Items[2].Caption := 'Links';

  LinkAwaitsLinkingAllForms();
  Result := ShowModal();
end;

procedure TSourcePinForm.UpdateLinksCount();
var
  Counter : Integer;
  I       : Integer;
begin
  Counter := 0;
  for I := 0 to ListView.Items.Count - 1 do
    if ( ListView.Items.Item[ i ].Checked ) then
      Inc( Counter );

  LinksCountLabel.Caption := IntToStr( Counter );
end;

procedure TSourcePinForm.ClearData();
begin
  ListView.Items.Clear();
end;

procedure TSourcePinForm.ListViewChange(Sender: TObject; Item: TListItem;
  Change: TItemChange);

var
  I     : Integer;
  Entry : TOWPinEntry;
  
begin
  if( ListUpdating ) then
    Exit;
    
  if ( Item.Checked and ( TImageIndex(Item.Data) = -1 )) then
    begin
    Item.Checked := False;
    Item.StateIndex := -1;
    end;

  if ( Item.StateIndex <> -1 ) then
    if ( Item.StateIndex <> Ord(Item.Checked )) then
      Item.StateIndex := Ord(Item.Checked );

  if( ( SinkPin <> NIL ) AND ( Item.Checked )) then
    for I := 0 to ListView.Items.Count - 1 do
      if( ListView.Items[ I ] <> Item ) then
        if( ListView.Items[ I ].Checked ) then
          ListView.Items[ I ].Checked := False;

  Entry := TOWPinEntry( Item.SubItems.Objects[ 1 ] );
  if( Entry <> NIL ) then
    Entry.Checked := Item.Checked;
    
  UpdateLinksCount();
end;

procedure TSourcePinForm.ListViewColumnClick(Sender: TObject; Column: TListColumn);
begin
  ListView.Columns.Items [ ColumnToSort ].ImageIndex := -1;
  
  if ( ColumnToSort = Column.Index ) then
    Direction := not Direction

  else
    begin
    Direction := true;
    ColumnToSort := Column.Index;
    end;

  TCustomListView(Sender).AlphaSort();

  ListView.Columns.Items [ ColumnToSort].ImageIndex := Integer( not Direction );
end;

procedure TSourcePinForm.FormCreate(Sender: TObject);
begin
  PinsList     := TObjectList.Create();
  PinsList.OwnsObjects := True;
  ColumnToSort := 0;
  Direction    := True;
  AllSelected  := False;

//  InspectorInvalidated := False;
end;


procedure TSourcePinForm.FormDestroy(Sender: TObject);
begin
  PinsList.Free();
  
end;

procedure TSourcePinForm.ListViewCompare(Sender: TObject; Item1, Item2: TListItem;
  Data: Integer; var Compare: Integer);
var
  DirMultiply : Integer;
  ix          : Integer;

begin

  if( Direction ) then
    DirMultiply :=  1
    
  else
    DirMultiply := -1;

  if (ColumnToSort = 0) then
    Compare := DirMultiply * CompareText ( Item1.Caption, Item2.Caption )

  else
    begin
    ix := ColumnToSort - 1;
    Compare := DirMultiply * CompareText ( Item1.SubItems.Strings[ix], Item2.SubItems.Strings[ix] );
    end

end;

function TSourcePinForm.GetListItems () : TListItems;
begin
  Result := ListView.Items;
end;
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
function  TOWPinListPropertyEditor.GetAttributes: TPropertyAttributes;
begin
  Result := [ paSubProperties, paReadOnly ];
  RequestRefresh();
end;
//---------------------------------------------------------------------------
procedure TOWPinListPropertyEditor.CheckRefresh();
var
  Collection : TOWPinList;
  
begin
  Collection := TOWPinList( GetOrdValue() );
  if( Collection.LastIndicatedCount <> Collection.Count ) then
    begin
    Collection.LastIndicatedCount := Collection.Count;
    ResetObjectInspector( Designer );
    end;
    
end;
//---------------------------------------------------------------------------
function  TOWPinListPropertyEditor.GetValue: string;
var
  Collection : TOWPinList;
  
begin
  RequestRefreshEx(Designer);
  
  OWLinkAwaitsLinking( Designer.Form );
  
  Collection := TOWPinList( GetOrdValue() );
  if( Collection.LastIndicatedCount <> Collection.Count ) then
    begin
    Collection.LastIndicatedCount := Collection.Count;
    ResetObjectInspector( Designer );
    end;

  if( Collection.Count = 0 ) then
    begin
    Result := '(Empty)';
    Exit;
    end;

  if( Collection.Count = 1 ) then
    begin
    Result := '1 Pin';
    Exit;
    end;

  Result := IntToStr ( Collection.Count ) + ' Pins';
end;
//---------------------------------------------------------------------------
procedure TOWPinListPropertyEditor.GetProperties(Proc: TAGetPropProc);
var
  Collection    : TOWPinList;
  I             : Integer;
  Pin           : TOWPin;
begin
  Collection := TOWPinList( GetOrdValue() );

  Collection.LastIndicatedCount := Collection.Count;
  for I := 0 to Collection.Count - 1 do
    begin
    try
      Pin := Collection.Pins[ I ];
      if( Pin is TOWSourcePin ) then
        Proc( TOWSourcePinListPropertyEditor.CreateEx( Designer, TOWSourcePin( Pin ), Collection.Names[ I ], Self ))

      else
        if( Pin is TOWSinkPin ) then
          Proc( TOWSinkPinListPropertyEditor.CreateEx( Designer, TOWSinkPin( Pin ), Collection.Names[ I ], Self ))

    finally
    end;

    end;


end;
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
function  TOWPinListOwnerPropertyEditor.GetAttributes: TPropertyAttributes;
begin
  Result := [ paSubProperties ];
  RequestRefresh();
end;
//---------------------------------------------------------------------------
function  TOWPinListOwnerPropertyEditor.GetValue: string;
var
  Collection : TOWPinList;

begin
  RequestRefreshEx(Designer);
  OWLinkAwaitsLinking( Designer.Form );
  Collection := TOWPinList( GetOrdValue() );
  Result := IntToStr ( Collection.Count );
end;
//---------------------------------------------------------------------------
procedure TOWPinListOwnerPropertyEditor.SetValue(const Value: string);
var
  Collection    : TOWPinList;

begin
  try
    Collection := TOWPinList( GetOrdValue() );
    Collection.Count := StrToInt( Value );
    Modified();
    ResetObjectInspector( Designer );
      
  except;
  end;
end;
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
function TSourcePinForm.RootFromName( RootName : String ) : TComponent;
var
  RealName : String;
  iPos     : Integer;

  CurrentComp : TComponent;

begin
  RealName := RootName;
  iPos := Pos( ' ', RealName );
  if( iPos > 0 ) then
    Delete( RealName, iPos, 10000 );

  Result := Designer.GetComponent( RealName + SEPARATOR + RealName );

  CurrentComp := Result; 
  while( CurrentComp.Owner <> NIL ) do
    begin
    if( Result is TDataModule ) then
      Break;
      
    if( ( CurrentComp.Owner is TCustomForm ) or ( CurrentComp.Owner is TDataModule ) ) then
      Result := CurrentComp.Owner;

    CurrentComp := CurrentComp.Owner
    end;

  if( Result = NIL ) then
    Result := Root;

end;
//---------------------------------------------------------------------------
procedure TSourcePinForm.FormsComboBoxChange(Sender: TObject);
var
  SelectedRoot : TComponent;
  FormName     : String;

begin
  SelectedRoot := NIL;
  AllSelected := False;
  if( FormsComboBox.ItemIndex = FormsComboBox.Items.Count - 1 ) then
    begin
    PopulateAll();
    AllSelected := True;
    Exit;
    end;

  if( FormsComboBox.ItemIndex <> -1 ) then
    begin
    FormName := FormsComboBox.Items.Strings[ FormsComboBox.ItemIndex ];
    SelectedRoot := RootFromName( FormName );
    end;

  PopulateForm( SelectedRoot );
end;


type TISlotsEditorNotifier = class( TIAddInNotifier )
    procedure FileNotification(NotifyCode: TFileNotification;
      const FileName: string; var Cancel: Boolean); override; stdcall;
    procedure EventNotification(NotifyCode: TEventNotification;
      var Cancel: Boolean); override; stdcall;
end;

procedure TISlotsEditorNotifier.FileNotification(NotifyCode: TFileNotification;
      const FileName: string; var Cancel: Boolean);
begin
  if( NotifyCode <> fnFileOpened ) then
    Exit;

  RequestRefresh();
end;

procedure TISlotsEditorNotifier.EventNotification(NotifyCode: TEventNotification;
      var Cancel: Boolean);
begin
end;

var EditorNotifier : TISlotsEditorNotifier;


procedure TSourcePinForm.Button1Click(Sender: TObject);
begin
  OWReportAwaitsLinking();
end;

initialization
  InOppening := False;
  InRefresh := False;
  EmptyList := TADesignerSelectionList.Create;
  Form := TSourcePinForm.Create ( NIL );

  EditorNotifier := TISlotsEditorNotifier.Create();
  ToolServices.AddNotifier( EditorNotifier );

finalization
  ToolServices.RemoveNotifier( EditorNotifier );
  EditorNotifier.Free();
  Form.Free;
  EmptyList.Free;

end.
